@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2020 The Fluent Bit Authors
+ *  Copyright (C) 2019-2021 The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -74,6 +74,7 @@ struct flb_kube {
     int api_port;
     int api_https;
     int use_journal;
+    int cache_use_docker_id;
     int labels;
     int annotations;
     int dummy_meta;
@@ -147,7 +148,7 @@ struct flb_kube {
     int dns_retries;
     int dns_wait_time;
 
-    struct flb_tls tls;
+    struct flb_tls *tls;
     struct flb_config *config;
     struct flb_hash *hash_table;
     struct flb_upstream *upstream;
