@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2020 The Fluent Bit Authors
+ *  Copyright (C) 2019-2021 The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -284,7 +284,7 @@ static char *influxdb_format(const char *tag, int tag_len,
                 influxdb_bulk_append_bulk(bulk, bulk_body, ' ') != 0) {
                 goto error;
             }
-        } 
+        }
         else {
             flb_plg_warn(ctx->ins, "skip send record, "
                          "since no record available "
@@ -419,13 +419,14 @@ static int cb_influxdb_init(struct flb_output_instance *ins, struct flb_config *
                                    ins->host.name,
                                    ins->host.port,
                                    io_flags,
-                                   &ins->tls);
+                                   ins->tls);
     if (!upstream) {
         flb_free(ctx);
         return -1;
     }
     ctx->u   = upstream;
     ctx->seq = 0;
+    flb_output_upstream_set(ctx->u, ins);
 
     flb_time_zero(&ctx->ts_dupe);
     flb_time_zero(&ctx->ts_last);
