@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2020 The Fluent Bit Authors
+ *  Copyright (C) 2019-2021 The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -281,16 +281,16 @@ static void s3_context_destroy(struct flb_s3 *ctx)
         flb_aws_provider_destroy(ctx->provider);
     }
 
-    if (ctx->provider_tls.context) {
-        flb_tls_context_destroy(ctx->provider_tls.context);
+    if (ctx->provider_tls) {
+        flb_tls_destroy(ctx->provider_tls);
     }
 
-    if (ctx->sts_provider_tls.context) {
-        flb_tls_context_destroy(ctx->sts_provider_tls.context);
+    if (ctx->sts_provider_tls) {
+        flb_tls_destroy(ctx->sts_provider_tls);
     }
 
-    if (ctx->client_tls.context) {
-        flb_tls_context_destroy(ctx->client_tls.context);
+    if (ctx->client_tls) {
+        flb_tls_destroy(ctx->client_tls);
     }
 
     if (ctx->s3_client) {
@@ -348,6 +348,16 @@ static int cb_s3_init(struct flb_output_instance *ins,
         return -1;
     }
 
+    /* Date key */
+    ctx->date_key = ctx->json_date_key;
+    tmp = flb_output_get_property("json_date_key", ins);
+    if (tmp) {
+        /* Just check if we have to disable it */
+        if (flb_utils_bool(tmp) == FLB_FALSE) {
+            ctx->date_key = NULL;
+        }
+    }
+
     /* Date format for JSON output */
     ctx->json_date_format = FLB_PACK_JSON_DATE_ISO8601;
     tmp = flb_output_get_property("json_date_format", ins);
@@ -502,35 +512,35 @@ static int cb_s3_init(struct flb_output_instance *ins,
         ctx->content_type = (char *) tmp;
     }
     
-    ctx->client_tls.context = flb_tls_context_new(FLB_TRUE,
-                                                  ins->tls_debug,
-                                                  ins->tls_vhost,
-                                                  ins->tls_ca_path,
-                                                  ins->tls_ca_file,
-                                                  ins->tls_crt_file,
-                                                  ins->tls_key_file,
-                                                  ins->tls_key_passwd);
-    if (!ctx->client_tls.context) {
+    ctx->client_tls = flb_tls_create(FLB_TRUE,
+                                     ins->tls_debug,
+                                     ins->tls_vhost,
+                                     ins->tls_ca_path,
+                                     ins->tls_ca_file,
+                                     ins->tls_crt_file,
+                                     ins->tls_key_file,
+                                     ins->tls_key_passwd);
+    if (!ctx->client_tls) {
         flb_plg_error(ctx->ins, "Failed to create tls context");
         return -1;
     }
 
     /* AWS provider needs a separate TLS instance */
-    ctx->provider_tls.context = flb_tls_context_new(FLB_TRUE,
-                                                    ins->tls_debug,
-                                                    ins->tls_vhost,
-                                                    ins->tls_ca_path,
-                                                    ins->tls_ca_file,
-                                                    ins->tls_crt_file,
-                                                    ins->tls_key_file,
-                                                    ins->tls_key_passwd);
-    if (!ctx->provider_tls.context) {
+    ctx->provider_tls = flb_tls_create(FLB_TRUE,
+                                       ins->tls_debug,
+                                       ins->tls_vhost,
+                                       ins->tls_ca_path,
+                                       ins->tls_ca_file,
+                                       ins->tls_crt_file,
+                                       ins->tls_key_file,
+                                       ins->tls_key_passwd);
+    if (!ctx->provider_tls) {
         flb_errno();
         return -1;
     }
 
     ctx->provider = flb_standard_chain_provider_create(config,
-                                                       &ctx->provider_tls,
+                                                       ctx->provider_tls,
                                                        ctx->region,
                                                        ctx->sts_endpoint,
                                                        NULL,
@@ -552,16 +562,16 @@ static int cb_s3_init(struct flb_output_instance *ins,
         }
 
         /* STS provider needs yet another separate TLS instance */
-        ctx->sts_provider_tls.context = flb_tls_context_new(FLB_TRUE,
-                                                            ins->tls_debug,
-                                                            ins->tls_vhost,
-                                                            ins->tls_ca_path,
-                                                            ins->tls_ca_file,
-                                                            ins->tls_crt_file,
-                                                            ins->tls_key_file,
-                                                            ins->tls_key_passwd);
+        ctx->sts_provider_tls = flb_tls_create(FLB_TRUE,
+                                               ins->tls_debug,
+                                               ins->tls_vhost,
+                                               ins->tls_ca_path,
+                                               ins->tls_ca_file,
+                                               ins->tls_crt_file,
+                                               ins->tls_key_file,
+                                               ins->tls_key_passwd);
 
-        if (!ctx->sts_provider_tls.context) {
+        if (!ctx->sts_provider_tls) {
             flb_errno();
             return -1;
         }
@@ -575,7 +585,7 @@ static int cb_s3_init(struct flb_output_instance *ins,
         }
 
         ctx->provider = flb_sts_provider_create(config,
-                                                &ctx->sts_provider_tls,
+                                                ctx->sts_provider_tls,
                                                 ctx->base_provider,
                                                 external_id,
                                                 role_arn,
@@ -630,7 +640,7 @@ static int cb_s3_init(struct flb_output_instance *ins,
     ctx->s3_client->s3_mode = S3_MODE_SIGNED_PAYLOAD;
 
     ctx->s3_client->upstream = flb_upstream_create(config, ctx->endpoint, 443,
-                                                   FLB_IO_TLS, &ctx->client_tls);
+                                                   FLB_IO_TLS, ctx->client_tls);
     if (!ctx->s3_client->upstream) {
         flb_plg_error(ctx->ins, "Connection initialization error");
         return -1;
@@ -699,6 +709,9 @@ static int cb_s3_init(struct flb_output_instance *ins,
         ctx->s3_client->upstream->flags = async_flags;
     }
 
+    /* this is done last since in the previous block we make calls to AWS */
+    ctx->provider->provider_vtable->upstream_set(ctx->provider, ctx->ins);
+
     return 0;
 }
 
@@ -1281,17 +1294,18 @@ static void cb_s3_flush(const void *data, size_t bytes,
                             void *out_context,
                             struct flb_config *config)
 {
-    struct flb_s3 *ctx = out_context;
-    flb_sds_t json = NULL;
-    struct s3_file *chunk = NULL;
-    struct multipart_upload *m_upload = NULL;
-    char *buffer = NULL;
-    size_t buffer_size;
-    int timeout_check = FLB_FALSE;
-    size_t chunk_size = 0;
-    size_t upload_size = 0;
     int ret;
     int len;
+    int timeout_check = FLB_FALSE;
+    char *buffer = NULL;
+    size_t buffer_size;
+    size_t chunk_size = 0;
+    size_t upload_size = 0;
+    flb_sds_t json = NULL;
+    struct s3_file *chunk = NULL;
+    struct flb_s3 *ctx = out_context;
+    struct multipart_upload *m_upload = NULL;
+    struct flb_sched *sched;
     (void) i_ins;
     (void) config;
 
@@ -1323,7 +1337,8 @@ static void cb_s3_flush(const void *data, size_t bytes,
                       "Creating upload timer with frequency %ds",
                       ctx->timer_ms / 1000);
 
-        ret = flb_sched_timer_cb_create(config, FLB_SCHED_TIMER_CB_PERM,
+        sched = flb_sched_ctx_get();
+        ret = flb_sched_timer_cb_create(sched, FLB_SCHED_TIMER_CB_PERM,
                                         ctx->timer_ms,
                                         cb_s3_upload,
                                         ctx);
@@ -1337,7 +1352,7 @@ static void cb_s3_flush(const void *data, size_t bytes,
     json = flb_pack_msgpack_to_json_format(data, bytes,
                                            FLB_PACK_JSON_FORMAT_LINES,
                                            ctx->json_date_format,
-                                           ctx->json_date_key);
+                                           ctx->date_key);
 
     if (json == NULL) {
         flb_plg_error(ctx->ins, "Could not marshal msgpack to JSON");
