@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2020 The Fluent Bit Authors
+ *  Copyright (C) 2019-2021 The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -35,7 +35,6 @@
 #include <fluent-bit/flb_plugin.h>
 #include <fluent-bit/flb_plugins.h>
 #include <fluent-bit/flb_slist.h>
-#include <fluent-bit/flb_io_tls.h>
 #include <fluent-bit/flb_kernel.h>
 #include <fluent-bit/flb_worker.h>
 #include <fluent-bit/flb_scheduler.h>
@@ -184,7 +183,7 @@ struct flb_config *flb_config_init()
 #endif
 
     /* Set default coroutines stack size */
-    config->coro_stack_size = FLB_THREAD_STACK_SIZE;
+    config->coro_stack_size = FLB_CORO_STACK_SIZE;
 
     /* Initialize linked lists */
     mk_list_init(&config->collectors);
@@ -244,7 +243,7 @@ void flb_config_exit(struct flb_config *config)
     }
 
     if (config->log) {
-        flb_log_stop(config->log, config);
+        flb_log_destroy(config->log, config);
     }
 
     if (config->parsers_file) {
@@ -316,6 +315,11 @@ void flb_config_exit(struct flb_config *config)
         flb_free(config->conf_path);
     }
 
+    /* Working directory */
+    if (config->workdir) {
+        flb_free(config->workdir);
+    }
+
     /* Destroy any DSO context */
     flb_plugin_destroy(config->dso_plugins);
 
@@ -329,7 +333,7 @@ void flb_config_exit(struct flb_config *config)
     mk_event_closesocket(config->flush_fd);
 
     /* Release scheduler */
-    flb_sched_exit(config);
+    flb_sched_destroy(config->sched);
 
 #ifdef FLB_HAVE_HTTP_SERVER
     if (config->http_listen) {
@@ -363,7 +367,6 @@ void flb_config_exit(struct flb_config *config)
         mk_event_loop_destroy(config->evl);
     }
 
-
     flb_plugins_unregister(config);
     flb_free(config);
 }
