@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2020 The Fluent Bit Authors
+ *  Copyright (C) 2019-2021 The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -25,6 +25,7 @@
 #include <fluent-bit/flb_time.h>
 #include <fluent-bit/flb_router.h>
 #include <fluent-bit/flb_task.h>
+#include <fluent-bit/flb_routes_mask.h>
 #include <fluent-bit/stream_processor/flb_sp.h>
 
 static void generate_chunk_name(struct flb_input_instance *in,
@@ -45,6 +46,17 @@ ssize_t flb_input_chunk_get_size(struct flb_input_chunk *ic)
     return cio_chunk_get_content_size(ic->chunk);
 }
 
+/*
+ * When chunk is set to DOWN from memory, data_size is set to 0 and
+ * cio_chunk_get_content_size(1) returns the data_size. fs_chunks_size
+ * is used to track the size of chunks in filesystem so we need to call
+ * cio_chunk_get_real_size to return the original size in the file system
+ */
+ssize_t flb_input_chunk_get_real_size(struct flb_input_chunk *ic)
+{
+    return cio_chunk_get_real_size(ic->chunk);
+}
+
 int flb_input_chunk_write(void *data, const char *buf, size_t len)
 {
     int ret;
@@ -94,7 +106,7 @@ static int flb_input_chunk_is_task_safe_delete(struct flb_task *task)
 
 int flb_input_chunk_safe_delete(struct flb_input_chunk *ic,
                                 struct flb_input_chunk *old_ic,
-                                uint64_t o_mask_id)
+                                uint64_t o_id)
 {
     /* The chunk we want to drop should not be the incoming chunk */
     if (ic == old_ic) {
@@ -104,9 +116,9 @@ int flb_input_chunk_safe_delete(struct flb_input_chunk *ic,
     /*
      * Even if chunks from same input plugin have same routes_mask when created,
      * the routes_mask could be modified when new chunks is ingested. Therefore,
-     * we still need to do the validation on the routes_mask with mask_id.
+     * we still need to do the validation on the routes_mask with o_id.
      */
-    if ((old_ic->routes_mask & o_mask_id) == 0) {
+    if (flb_routes_mask_get_bit(old_ic->routes_mask, o_id) == 0) {
         return FLB_FALSE;
     }
 
@@ -131,12 +143,12 @@ int flb_intput_chunk_count_dropped_chunks(struct flb_input_chunk *ic,
     mk_list_foreach(head, &ic->in->chunks) {
         old_ic = mk_list_entry(head, struct flb_input_chunk, _head);
 
-        if (flb_input_chunk_safe_delete(ic, old_ic, o_ins->mask_id) == FLB_FALSE ||
+        if (flb_input_chunk_safe_delete(ic, old_ic, o_ins->id) == FLB_FALSE ||
             flb_input_chunk_is_task_safe_delete(old_ic->task) == FLB_FALSE) {
             continue;
         }
 
-        bytes_remained += flb_input_chunk_get_size(old_ic);
+        bytes_remained += flb_input_chunk_get_real_size(old_ic);
         count++;
         if (bytes_remained >= chunk_size) {
             enough_space = FLB_TRUE;
@@ -162,12 +174,8 @@ int flb_intput_chunk_count_dropped_chunks(struct flb_input_chunk *ic,
 /*
  * Find a slot in the output instance to append the new data with size chunk_size, it
  * will drop the the oldest chunks when the limitation on local disk is reached.
- *
- * overlimit_routes_mask: A bit mask used to check whether the output instance will
- * reach the limit when buffering the new data
  */
 int flb_input_chunk_find_space_new_data(struct flb_input_chunk *ic,
-                                        uint64_t overlimit_routes_mask,
                                         size_t chunk_size)
 {
     int count;
@@ -188,7 +196,7 @@ int flb_input_chunk_find_space_new_data(struct flb_input_chunk *ic,
         o_ins = mk_list_entry(head, struct flb_output_instance, _head);
 
         if ((o_ins->total_limit_size == -1) ||
-            (ic->routes_mask & o_ins->mask_id) == 0) {
+            (flb_routes_mask_get_bit(ic->routes_mask, o_ins->id) == 0)) {
             continue;
         }
 
@@ -202,7 +210,7 @@ int flb_input_chunk_find_space_new_data(struct flb_input_chunk *ic,
              */
             flb_error("[input chunk] no enough space in filesystem to buffer "
                       "chunk %s in plugin %s", flb_input_chunk_get_name(ic), o_ins->name);
-            ic->routes_mask ^= o_ins->mask_id;
+            flb_routes_mask_clear_bit(ic->routes_mask, o_ins->id);
             continue;
         }
 
@@ -215,21 +223,22 @@ int flb_input_chunk_find_space_new_data(struct flb_input_chunk *ic,
         mk_list_foreach_safe(head_chunk, tmp, &ic->in->chunks) {
             old_ic = mk_list_entry(head_chunk, struct flb_input_chunk, _head);
 
-            if (flb_input_chunk_safe_delete(ic, old_ic, o_ins->mask_id) == FLB_FALSE ||
+            if (flb_input_chunk_safe_delete(ic, old_ic, o_ins->id) == FLB_FALSE ||
                 flb_input_chunk_is_task_safe_delete(old_ic->task) == FLB_FALSE) {
                 continue;
             }
 
-            old_ic_bytes = flb_input_chunk_get_size(old_ic);
+            old_ic_bytes = flb_input_chunk_get_real_size(old_ic);
+
             /* drop chunk by adjusting the routes_mask */
-            old_ic->routes_mask ^= o_ins->mask_id;
+            flb_routes_mask_clear_bit(old_ic->routes_mask, o_ins->id);
             o_ins->fs_chunks_size -= old_ic_bytes;
 
             flb_debug("[input chunk] remove route of chunk %s with size %ld bytes to output plugin %s "
                       "to place the incoming data with size %ld bytes", flb_input_chunk_get_name(old_ic),
                       old_ic_bytes, o_ins->name, chunk_size);
 
-            if (old_ic->routes_mask == 0) {
+            if (flb_routes_mask_is_empty(old_ic->routes_mask)) {
                 if (old_ic->task != NULL) {
                     /*
                      * If the chunk is referenced by a task and task has no active route,
@@ -260,13 +269,13 @@ int flb_input_chunk_find_space_new_data(struct flb_input_chunk *ic,
 }
 
 /*
- * Returns routes_mask of output instances that will reach the limit
+ * Returns a non-zero result if any output instances will reach the limit
  * after buffering the new data
  */
-uint64_t flb_input_chunk_get_overlimit_routes_mask(struct flb_input_chunk *ic,
-                                                   size_t chunk_size)
+int flb_input_chunk_has_overlimit_routes(struct flb_input_chunk *ic,
+                                         size_t chunk_size)
 {
-    uint64_t routes_mask = 0;
+    int overlimit = 0;
     struct mk_list *head;
     struct flb_output_instance *o_ins;
 
@@ -274,43 +283,46 @@ uint64_t flb_input_chunk_get_overlimit_routes_mask(struct flb_input_chunk *ic,
         o_ins = mk_list_entry(head, struct flb_output_instance, _head);
 
         if ((o_ins->total_limit_size == -1) ||
-            (ic->routes_mask & o_ins->mask_id) == 0) {
+            (flb_routes_mask_get_bit(ic->routes_mask, o_ins->id) == 0)) {
             continue;
         }
 
-        flb_debug("[input chunk] chunk %s required %ld bytes and %ld bytes left in plugin %s",
+        flb_debug("[input chunk] chunk %s required %ld bytes and %ld bytes left "
+                  "in plugin %s",
                   flb_input_chunk_get_name(ic), chunk_size,
                   o_ins->total_limit_size - o_ins->fs_chunks_size,
                   o_ins->name);
 
         if (o_ins->fs_chunks_size + chunk_size > o_ins->total_limit_size) {
-            routes_mask |= o_ins->mask_id;
+            overlimit = 1;
         }
     }
 
-    return routes_mask;
+    return overlimit;
 }
 
-/* Find a slot for the incoming data to buffer it in local file system */
-uint64_t flb_input_chunk_place_new_chunk(struct flb_input_chunk *ic, size_t chunk_size)
+/* Find a slot for the incoming data to buffer it in local file system
+ * returns 0 if none of the routes can be written to
+ */
+int flb_input_chunk_place_new_chunk(struct flb_input_chunk *ic, size_t chunk_size)
 {
-    uint64_t overlimit_routes_mask;
-    overlimit_routes_mask = flb_input_chunk_get_overlimit_routes_mask(ic, chunk_size);
-    if (overlimit_routes_mask != 0) {
-        flb_input_chunk_find_space_new_data(ic, overlimit_routes_mask, chunk_size);
+	int overlimit;
+    overlimit = flb_input_chunk_has_overlimit_routes(ic, chunk_size);
+    if (overlimit != 0) {
+        flb_input_chunk_find_space_new_data(ic, chunk_size);
     }
 
-    return ic->routes_mask;
+    return !flb_routes_mask_is_empty(ic->routes_mask);
 }
 
 /* Create an input chunk using a Chunk I/O */
 struct flb_input_chunk *flb_input_chunk_map(struct flb_input_instance *in,
                                             void *chunk)
 {
-    uint64_t chunk_routes_mask;
     ssize_t bytes;
     const char *tag_buf;
     int tag_len;
+    int has_routes;
     int ret;
 
 #ifdef FLB_HAVE_METRICS
@@ -320,7 +332,7 @@ struct flb_input_chunk *flb_input_chunk_map(struct flb_input_instance *in,
     struct flb_input_chunk *ic;
 
     /* Create context for the input instance */
-    ic = flb_malloc(sizeof(struct flb_input_chunk));
+    ic = flb_calloc(1, sizeof(struct flb_input_chunk));
     if (!ic) {
         flb_errno();
         return NULL;
@@ -331,13 +343,13 @@ struct flb_input_chunk *flb_input_chunk_map(struct flb_input_instance *in,
     ic->chunk = chunk;
     ic->in = in;
     msgpack_packer_init(&ic->mp_pck, ic, flb_input_chunk_write);
-    mk_list_add(&ic->_head, &in->chunks);
 
 #ifdef FLB_HAVE_METRICS
     ret = cio_chunk_get_content(ic->chunk, &buf_data, &buf_size);
     if (ret != CIO_OK) {
         flb_error("[input chunk] error retrieving content for metrics");
-        return ic;
+        flb_free(ic);
+        return NULL;
     }
 
     ic->total_records = flb_mp_count(buf_data, buf_size);
@@ -350,18 +362,25 @@ struct flb_input_chunk *flb_input_chunk_map(struct flb_input_instance *in,
     /* Get the the tag reference (chunk metadata) */
     ret = flb_input_chunk_get_tag(ic, &tag_buf, &tag_len);
     if (ret == -1) {
-	flb_error("[input chunk] error retrieving tag of input chunk");
-	return ic;
+        flb_error("[input chunk] error retrieving tag of input chunk");
+        flb_free(ic);
+        return NULL;
     }
 
-    chunk_routes_mask = flb_router_get_routes_mask_by_tag(tag_buf, tag_len, in);
-    if (chunk_routes_mask == 0) {
+    bytes = flb_input_chunk_get_real_size(ic);
+    if (bytes < 0) {
+        flb_warn("[input chunk] could not retrieve chunk real size");
+        flb_free(ic);
+        return NULL;
+    }
+
+    has_routes = flb_routes_mask_set_by_tag(ic->routes_mask, tag_buf, tag_len, in);
+    if (has_routes == 0) {
         flb_warn("[input chunk] no matching route for backoff log chunk %s",
                  flb_input_chunk_get_name(ic));
     }
-    ic->routes_mask = chunk_routes_mask;
 
-    bytes = flb_input_chunk_get_size(ic);
+    mk_list_add(&ic->_head, &in->chunks);
     flb_input_chunk_update_output_instances(ic, bytes);
 
     return ic;
@@ -373,7 +392,7 @@ struct flb_input_chunk *flb_input_chunk_create(struct flb_input_instance *in,
     int ret;
     int err;
     int set_down = FLB_FALSE;
-    uint64_t chunk_routes_mask;
+    int has_routes;
     char name[64];
     struct cio_chunk *chunk;
     struct flb_storage_input *storage;
@@ -439,12 +458,11 @@ struct flb_input_chunk *flb_input_chunk_create(struct flb_input_instance *in,
 #endif
 
     /* Calculate the routes_mask for the input chunk */
-    chunk_routes_mask = flb_router_get_routes_mask_by_tag(tag, tag_len, in);
-    if (chunk_routes_mask == 0) {
+    has_routes = flb_routes_mask_set_by_tag(ic->routes_mask, tag, tag_len, in);
+    if (has_routes == 0) {
         flb_trace("[input chunk] no matching route for input chunk '%s' with tag '%s'",
                   flb_input_chunk_get_name(ic), tag);
     }
-    ic->routes_mask = chunk_routes_mask;
 
     msgpack_packer_init(&ic->mp_pck, ic, flb_input_chunk_write);
     mk_list_add(&ic->_head, &in->chunks);
@@ -453,15 +471,23 @@ struct flb_input_chunk *flb_input_chunk_create(struct flb_input_instance *in,
         cio_chunk_down(chunk);
     }
 
+    flb_hash_add(in->ht_chunks, tag, tag_len, ic, 0);
     return ic;
 }
 
 int flb_input_chunk_destroy(struct flb_input_chunk *ic, int del)
 {
+    int tag_len;
+    int ret;
     ssize_t bytes;
+    const char *tag_buf = NULL;
     struct mk_list *head;
     struct flb_output_instance *o_ins;
 
+    if (flb_input_chunk_is_up(ic) == FLB_FALSE) {
+        flb_input_chunk_set_up(ic);
+    }
+
     mk_list_foreach(head, &ic->in->config->outputs) {
         o_ins = mk_list_entry(head, struct flb_output_instance, _head);
 
@@ -469,12 +495,43 @@ int flb_input_chunk_destroy(struct flb_input_chunk *ic, int del)
             continue;
         }
 
-        bytes = flb_input_chunk_get_size(ic);
-        if ((ic->routes_mask & o_ins->mask_id) > 0) {
+        bytes = flb_input_chunk_get_real_size(ic);
+        if (flb_routes_mask_get_bit(ic->routes_mask, o_ins->id) != 0) {
             o_ins->fs_chunks_size -= bytes;
         }
     }
 
+    /*
+     * When a chunk is going to be destroyed, this can be in a down state,
+     * since the next step is to retrieve the Tag we need to have the
+     * content up.
+     */
+    ret = flb_input_chunk_is_up(ic);
+    if (ret == FLB_FALSE) {
+        ret = cio_chunk_up_force(ic->chunk);
+        if (ret == -1) {
+            flb_error("[input chunk] cannot load chunk: %s",
+                      flb_input_chunk_get_name(ic));
+        }
+    }
+
+    /* Retrieve Tag */
+    ret = flb_input_chunk_get_tag(ic, &tag_buf, &tag_len);
+    if (ret == -1) {
+        flb_trace("[input chunk] could not retrieve chunk tag: %s",
+                  flb_input_chunk_get_name(ic));
+    }
+
+    if (del == CIO_TRUE && tag_buf) {
+        /*
+         * "TRY" to delete any reference to this chunk ('ic') from the hash
+         * table. Note that maybe the value is not longer available in the
+         * entries if it was replaced: note that we always keep the last
+         * chunk for a specific Tag.
+         */
+        flb_hash_del_ptr(ic->in->ht_chunks, tag_buf, tag_len, (void *) ic);
+    }
+
     cio_chunk_close(ic->chunk, del);
     mk_list_del(&ic->_head);
     flb_free(ic);
@@ -487,28 +544,23 @@ static struct flb_input_chunk *input_chunk_get(const char *tag, int tag_len,
                                                struct flb_input_instance *in,
                                                size_t chunk_size)
 {
+    int id;
+    int ret;
     int new_chunk = FLB_FALSE;
-    struct mk_list *head;
+    size_t out_size;
     struct flb_input_chunk *ic = NULL;
 
-    /* Try to find a current chunk context to append the data */
-    mk_list_foreach_r(head, &in->chunks) {
-        ic = mk_list_entry(head, struct flb_input_chunk, _head);
+    id = flb_hash_get(in->ht_chunks, tag, tag_len, (void *) &ic, &out_size);
+    if (id >= 0) {
         if (ic->busy == FLB_TRUE || cio_chunk_is_locked(ic->chunk)) {
             ic = NULL;
-            continue;
         }
-
-        if (cio_chunk_is_up(ic->chunk) == CIO_FALSE) {
-            ic = NULL;
-            continue;
+        else if (cio_chunk_is_up(ic->chunk) == CIO_FALSE) {
+            ret = cio_chunk_up_force(ic->chunk);
+            if (ret == -1) {
+                ic = NULL;
+            }
         }
-
-        if (cio_meta_cmp(ic->chunk, (char *) tag, tag_len) != 0) {
-            ic = NULL;
-            continue;
-        }
-        break;
     }
 
     /* No chunk was found, we need to create a new one */
@@ -525,8 +577,8 @@ static struct flb_input_chunk *input_chunk_get(const char *tag, int tag_len,
      * that the chunk will flush to, we need to modify the routes_mask of the oldest chunks
      * (based in creation time) to get enough space for the incoming chunk.
      */
-    if (ic->routes_mask != 0 &&
-        flb_input_chunk_place_new_chunk(ic, chunk_size) == 0) {
+    if (!flb_routes_mask_is_empty(ic->routes_mask)
+        && flb_input_chunk_place_new_chunk(ic, chunk_size) == 0) {
         /*
          * If the chunk is not newly created, the chunk might already have logs inside.
          * We cannot delete (reused) chunks here.
@@ -595,7 +647,6 @@ size_t flb_input_chunk_set_limits(struct flb_input_instance *in)
 
     /* Gather total number of enqueued bytes */
     total = flb_input_chunk_total_size(in);
-
     /* Register the total into the context variable */
     in->mem_chunks_size = total;
 
@@ -963,7 +1014,7 @@ void flb_input_chunk_update_output_instances(struct flb_input_chunk *ic,
             continue;
         }
 
-        if ((ic->routes_mask & o_ins->mask_id) > 0) {
+        if (flb_routes_mask_get_bit(ic->routes_mask, o_ins->id) != 0) {
             /*
              * if there is match on any index of 1's in the binary, it indicates
              * that the input chunk will flush to this output instance
