@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2020 The Fluent Bit Authors
+ *  Copyright (C) 2019-2021 The Fluent Bit Authors
  *  Copyright (C) 2015-2018 Treasure Data Inc.
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
@@ -434,10 +434,10 @@ int64_t flb_utils_size_to_bytes(const char *size)
     return val;
 }
 
-int flb_utils_hex2int(char *hex, int len)
+int64_t flb_utils_hex2int(char *hex, int len)
 {
     int i = 0;
-    int res = 0;
+    int64_t res = 0;
     char c;
 
     while ((c = *hex++) && i < len) {
@@ -931,6 +931,7 @@ int flb_utils_proxy_url_split(const char *in_url, char **out_protocol,
         /* Parse username:passwrod part. */
         tmp = strchr(proto_sep, ':');
         if (!tmp) {
+            flb_free(protocol);
             return -1;
         }
         username = mk_string_copy_substr(proto_sep, 0, tmp - proto_sep);
